import pytest

import abjad


def test_obgc_01():
    """
    Raises exception when duration of on-beat grace container exceeds duration
    of anchor container.
    """

    music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
    string = "g'8 a' b' c'' d'' c'' b' a' b' c'' d''"
    with pytest.raises(Exception):
        abjad.on_beat_grace_container(
            string,
            music_voice[1:2],
            grace_leaf_duration=(1, 8),
        )



import dataclasses

from . import _indentlib
from . import configuration as _configuration
from . import indicators as _indicators
from . import iterate as _iterate
from . import score as _score
from . import tag as _tag

configuration = _configuration.Configuration()


@dataclasses.dataclass
class Block:
    r'''
    LilyPond file block.

    ..  container:: example

        Use strings to add contents to a block:

        >>> string = r"""right-margin = 2\cm
        ...     left-margin = 2\cm"""
        >>> block = abjad.Block("paper", items=[string])
        >>> string = abjad.lilypond(block)
        >>> print(string)
        \paper
        {
            right-margin = 2\cm
            left-margin = 2\cm
        }

    ..  container:: example

        Define a context block like this:

        >>> string = r"""\Staff
        ...     \name FluteStaff
        ...     \type Engraver_group
        ...     \alias Staff
        ...     \remove Forbid_line_break_engraver
        ...     \consists Horizontal_bracket_engraver
        ...     \accepts FluteUpperVoice
        ...     \accepts FluteLowerVoice
        ...     \override Beam.positions = #'(-4 . -4)
        ...     \override Stem.stem-end-position = -6
        ...     autoBeaming = ##f
        ...     tupletFullLength = ##t
        ...     \accidentalStyle dodecaphonic"""
        >>> block = abjad.Block("context", items=[string])
        >>> string = abjad.lilypond(block)
        >>> print(string)
        \context
        {
            \Staff
            \name FluteStaff
            \type Engraver_group
            \alias Staff
            \remove Forbid_line_break_engraver
            \consists Horizontal_bracket_engraver
            \accepts FluteUpperVoice
            \accepts FluteLowerVoice
            \override Beam.positions = #'(-4 . -4)
            \override Stem.stem-end-position = -6
            autoBeaming = ##f
            tupletFullLength = ##t
            \accidentalStyle dodecaphonic
        }

    ..  container:: example

        Define an anonymous block like this:

        >>> string = r"""command-1
        ...     command-2
        ...     command-3"""
        >>> block = abjad.Block("", items=[string])
        >>> string = abjad.lilypond(block)
        >>> print(string)
        {
            command-1
            command-2
            command-3
        }

    ..  container:: example

        Markup formats like this:

        >>> block = abjad.Block("score", items=[abjad.Markup(r"\markup foo")])
        >>> string = abjad.lilypond(block)
        >>> print(string)
        \score
        {
            \markup foo
        }

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> block = abjad.Block("score")
        >>> block.items.append("<<")
        >>> block.items.append(r'{ \include "layout.ly" }')
        >>> block.items.append(staff)
        >>> block.items.append(">>")
        >>> lilypond_file = abjad.LilyPondFile(
        ...     lilypond_language_token=False,
        ...     lilypond_version_token=False,
        ... )
        >>> lilypond_file.items.append(block)
        >>> string = abjad.lilypond(lilypond_file)
        >>> print(string)
        \score
        {
            <<
            { \include "layout.ly" }
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }
            >>
        }

    '''

    name: str | None = None
    items: list = dataclasses.field(default_factory=list)

    def __post_init__(self):
        assert isinstance(self.items, list), repr(self.items)

    @staticmethod
    def _format_item(item, depth=1):
        indent = depth * _indentlib.INDENT
        result = []
        if isinstance(item, str):
            if item.isspace():
                string = ""
            else:
                string = indent + item
            result.append(string)
        elif hasattr(item, "_get_lilypond_format"):
            string = item._get_lilypond_format()
            pieces = string.split("\n")
            for piece in pieces:
                if piece.isspace():
                    piece = ""
                else:
                    piece = indent + piece
                result.append(piece)
        else:
            assert isinstance(item, _indicators.Markup), repr(item)
            string = indent + item.string
            result.append(string)
        return result

    def _get_lilypond_format(self, tag=None):
        result = []
        if not len(self.items):
            if self.name:
                string = rf"\{self.name} {{}}"
            else:
                string = "{}"
            return string
        strings = []
        if self.name:
            strings.append(rf"\{self.name}")
        strings.append("{")
        if tag is not None:
            strings = _tag.double_tag(strings, tag)
        result.extend(strings)
        for item in self.items:
            result.extend(self._format_item(item))
        string = "}"
        strings = [string]
        if tag is not None:
            strings = _tag.double_tag(strings, tag)
        result.extend(strings)
        string = "\n".join(result)
        return string


@dataclasses.dataclass(slots=True)
class LilyPondFile:
    r"""
    LilyPond file.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> lilypond_file = abjad.LilyPondFile(
        ...     items=[
        ...         r'\include "abjad.ily"',
        ...         '''#(set-default-paper-size "a5" 'portrait)''',
        ...         "#(set-global-staff-size 16)",
        ...         staff,
        ...     ],
        ... )

        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ::

            >>> string = abjad.lilypond(lilypond_file)
            >>> print(string)
            \version "..."
            \language "english"
            \include "abjad.ily"
            #(set-default-paper-size "a5" 'portrait)
            #(set-global-staff-size 16)
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
            }

    """

    items: list = dataclasses.field(default_factory=list)
    lilypond_language_token: bool | str = True
    lilypond_version_token: bool | str = True
    tag: _tag.Tag | None = None

    def __contains__(self, argument) -> bool:
        """
        Is true when LilyPond file contains ``argument``.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'", name="Staff")
            >>> lilypond_file = abjad.LilyPondFile([staff])

            >>> "Staff" in lilypond_file
            True

            >>> "Allegro" in lilypond_file
            False

            >>> 0 in lilypond_file
            False

        """
        try:
            self[argument]
            return True
        except (AssertionError, KeyError, ValueError, TypeError):
            return False

    def __getitem__(self, argument):
        r"""
        Gets item identified by ``argument``.

        ..  container:: example

            Searches score:

            >>> voice_1 = abjad.Voice("c''4 b' a' g'", name="Voice_1")
            >>> command = abjad.VoiceNumber(1)
            >>> abjad.attach(command, voice_1[0])
            >>> voice_2 = abjad.Voice("c'4 d' e' f'", name="Voice_2")
            >>> command = abjad.VoiceNumber(2)
            >>> abjad.attach(command, voice_2[0])
            >>> staff = abjad.Staff(
            ...     [voice_1, voice_2],
            ...     simultaneous=True,
            ...     name="Staff",
            ... )
            >>> score = abjad.Score([staff], name="Score")
            >>> block = abjad.Block("score")
            >>> block.items.append(score)
            >>> lilypond_file = abjad.LilyPondFile([block])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \context Score = "Score"
                <<
                    \context Staff = "Staff"
                    <<
                        \context Voice = "Voice_1"
                        {
                            \voiceOne
                            c''4
                            b'4
                            a'4
                            g'4
                        }
                        \context Voice = "Voice_2"
                        {
                            \voiceTwo
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    >>
                >>

            >>> lilypond_file["score"]
            Block(name='score', items=[Score("{ { c''4 b'4 a'4 g'4 } { c'4 d'4 e'4 f'4 } }", name='Score', simultaneous=True)])

            >>> lilypond_file["Score"]
            Score("{ { c''4 b'4 a'4 g'4 } { c'4 d'4 e'4 f'4 } }", name='Score', simultaneous=True)

            >>> lilypond_file["Staff"]
            Staff("{ c''4 b'4 a'4 g'4 } { c'4 d'4 e'4 f'4 }", name='Staff', simultaneous=True)

            >>> lilypond_file["Voice_1"]
            Voice("c''4 b'4 a'4 g'4", name='Voice_1')

            >>> lilypond_file["Voice_2"]
            Voice("c'4 d'4 e'4 f'4", name='Voice_2')

        ..  container:: example

            Searches score:

            >>> voice_1 = abjad.Voice("c''4 b' a' g'", name="Voice_1")
            >>> command = abjad.VoiceNumber(1)
            >>> abjad.attach(command, voice_1[0])
            >>> voice_2 = abjad.Voice("c'4 d' e' f'", name="Voice_2")
            >>> command = abjad.VoiceNumber(2)
            >>> abjad.attach(command, voice_2[0])
            >>> staff = abjad.Staff(
            ...     [voice_1, voice_2],
            ...     simultaneous=True,
            ...     name="Staff",
            ... )
            >>> score = abjad.Score([staff], name="Score")
            >>> lilypond_file = abjad.LilyPondFile([score])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \context Score = "Score"
                <<
                    \context Staff = "Staff"
                    <<
                        \context Voice = "Voice_1"
                        {
                            \voiceOne
                            c''4
                            b'4
                            a'4
                            g'4
                        }
                        \context Voice = "Voice_2"
                        {
                            \voiceTwo
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    >>
                >>

            >>> lilypond_file["Score"]
            Score("{ { c''4 b'4 a'4 g'4 } { c'4 d'4 e'4 f'4 } }", name='Score', simultaneous=True)

            >>> lilypond_file["Staff"]
            Staff("{ c''4 b'4 a'4 g'4 } { c'4 d'4 e'4 f'4 }", name='Staff', simultaneous=True)

            >>> lilypond_file["Voice_1"]
            Voice("c''4 b'4 a'4 g'4", name='Voice_1')

            >>> lilypond_file["Voice_2"]
            Voice("c'4 d'4 e'4 f'4", name='Voice_2')

        ..  container:: example

            REGRESSION. Works when score block contains parallel container:

            >>> include_container = abjad.Container()
            >>> string = r'\include "layout.ly"'
            >>> literal = abjad.LilyPondLiteral(string, site="opening")
            >>> abjad.attach(literal, include_container)
            >>> staff = abjad.Staff("c'4 d' e' f'", name="Staff")
            >>> container = abjad.Container(
            ...     [include_container, staff],
            ...     simultaneous=True,
            ... )
            >>> block = abjad.Block("score")
            >>> block.items.append(container)
            >>> lilypond_file = abjad.LilyPondFile(
            ...     items=[block],
            ...     lilypond_language_token=False,
            ...     lilypond_version_token=False,
            ... )
            >>> string = abjad.lilypond(lilypond_file)
            >>> print(string)
            \score
            {
                <<
                    {
                        \include "layout.ly"
                    }
                    \context Staff = "Staff"
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                >>
            }

            >>> lilypond_file['Staff']
            Staff("c'4 d'4 e'4 f'4", name='Staff')

        ..  container:: example

            Finds blocks by name:

            >>> blocks = [abjad.Block("header"), abjad.Block("score")]
            >>> lilypond_file = abjad.LilyPondFile(items=blocks)
            >>> lilypond_file["score"]
            Block(name='score', items=[])

            >>> score = abjad.Score([abjad.Staff("c'4 d' e' f'")], name="Score")
            >>> lilypond_file["score"].items.append(score)
            >>> string = abjad.lilypond(lilypond_file)
            >>> print(string)
            \version "..."
            \language "english"
            \header {}
            \score
            {
                \context Score = "Score"
                <<
                    \new Staff
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                >>
            }

        Returns block or component.
        """
        assert isinstance(argument, str), repr(argument)
        for item in self.items:
            if getattr(item, "name", None) == argument:
                return item
            elif hasattr(item, "items"):
                for item_ in item.items:
                    if getattr(item_, "name", None) == argument:
                        return item_
                    if isinstance(item_, _score.Component):
                        for component in _iterate.components(item_):
                            if getattr(component, "name", None) == argument:
                                return component
            elif isinstance(item, _score.Component):
                for component in _iterate.components(item):
                    if getattr(component, "name", None) == argument:
                        return component
        raise KeyError(f"no block or component with name {argument!r}.")

    def _get_lilypond_format(self):
        result = []
        strings = []
        if self.lilypond_version_token is True:
            string = configuration.get_lilypond_version_string()
            string = rf'\version "{string}"'
            strings.append(string)
        elif isinstance(self.lilypond_version_token, str):
            strings.append(self.lilypond_version_token)
        if self.lilypond_language_token is True:
            string = r'\language "english"'
            strings.append(string)
        # TODO: change to abjad.LilyPondFile._get_lilypond_format()
        tag = _tag.Tag("abjad.LilyPondFile._get_format_pieces()")
        tag = self.get_tag(tag)
        strings = _tag.double_tag(strings, tag)
        result.extend(strings)
        for item in self.items:
            if isinstance(item, str):
                result.append(item)
            else:
                try:
                    string = item._get_lilypond_format(tag=tag)
                except TypeError:
                    string = item._get_lilypond_format()
                assert isinstance(string, str), repr(string)
                result.append(string)
        strings = result
        string = "\n".join(strings)
        lines = []
        for line in string.split("\n"):
            if line.isspace():
                lines.append("")
            else:
                lines.append(line)
        return "\n".join(lines)

    def get_tag(self, site=None):
        """
        Gets tag.
        """
        if self.tag:
            tag = _tag.Tag(self.tag.string)
        else:
            tag = _tag.Tag()
        tag = tag.append(site)
        return tag



import copy
import enum
import importlib
import typing

from . import _getlib, _updatelib
from . import duration as _duration
from . import enums as _enums
from . import exceptions as _exceptions
from . import score as _score
from . import tag as _tag
from . import tweaks as _tweaks


def _before_attach(indicator, context, deactivate, component):
    if getattr(indicator, "temporarily_do_not_check", False) is True:
        return
    if hasattr(indicator, "allowable_sites"):
        if indicator.site not in component._allowable_sites:
            message = f"\n  {indicator!r}"
            message += f"\n  Can not attach to {component.__class__.__name__}."
            message += f"\n  {component.__class__.__name__} allows only "
            message += ", ".join(component._allowable_sites)
            message += " sites."
            raise Exception(message)
    if not hasattr(indicator, "context"):
        return
    if getattr(indicator, "find_context_on_attach", False) is True:
        the_context = context or indicator.context
        if the_context is not None:
            context_ = Wrapper._find_correct_effective_context(component, the_context)
            if context_ is None:
                message = f"\n    {indicator} requires {the_context} context;"
                message += f"\n    can not find {the_context}"
                message += f" in parentage of {component!r}."
                raise _exceptions.MissingContextError(message)
    if getattr(indicator, "nestable_spanner", False) is True:
        return
    if deactivate is True:
        return
    for wrapper in component._get_indicators(unwrap=False):
        if not isinstance(wrapper.unbundle_indicator(), type(indicator)):
            continue
        if getattr(indicator, "leak", None) != getattr(
            wrapper.unbundle_indicator(), "leak", None
        ):
            continue
        if indicator != wrapper.unbundle_indicator():
            if (
                getattr(indicator, "allow_multiple_with_different_values", False)
                is True
            ):
                continue
            if hasattr(indicator, "hide"):
                if indicator.hide != wrapper.unbundle_indicator().hide:
                    continue
            if getattr(indicator, "site", None) != getattr(
                wrapper.unbundle_indicator(), "site", None
            ):
                continue
        classname = type(component).__name__
        message = f"attempting to attach conflicting indicator to {classname}:"
        message += "\n  Already attached:"
        message += f"\n    {wrapper.unbundle_indicator()!r}"
        message += "\n  Attempting to attach:"
        message += f"\n    {indicator!r}"
        raise _exceptions.PersistentIndicatorError(message)


class Wrapper:
    r"""
    Wrapper.

    ..  container:: example

        >>> component = abjad.Note("c'4")
        >>> articulation = abjad.Articulation("accent")
        >>> abjad.attach(articulation, component, direction=abjad.UP)
        >>> abjad.get.wrapper(component)
        Wrapper(annotation=None, context=None, deactivate=False, direction=<Vertical.UP: 1>, indicator=Articulation(name='accent'), synthetic_offset=None, tag=Tag(string=''))

    ..  container:: example

        Duplicate indicator warnings take two forms.

        >>> voice_1 = abjad.Voice("c''4 d'' e'' f''", name="VoiceI")
        >>> voice_2 = abjad.Voice("c'4 d' e' f'", name="VoiceII")
        >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
        >>> abjad.attach(abjad.Clef("alto"), voice_2[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            <<
                \context Voice = "VoiceI"
                {
                    c''4
                    d''4
                    e''4
                    f''4
                }
                \context Voice = "VoiceII"
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

        First form when attempting to attach a contexted indicator to a leaf that already
        carries a contexted indicator of the same type:

        >>> # abjad.attach(abjad.Clef("treble"), voice_2[0])

        Second form when attempting to attach a contexted indicator to a leaf governed by
        some other component carrying a contexted indicator of the same type.

        >>> # abjad.attach(abjad.Clef("treble"), voice_1[0])

    """

    __documentation_section__ = "Internals"

    __slots__ = (
        "_annotation",
        "_component",
        "_context",
        "_deactivate",
        "_direction",
        "_effective_context",
        "_indicator",
        "_synthetic_offset",
        "_tag",
    )

    def __init__(
        self,
        annotation: str | enum.Enum | None = None,
        # TODO: maybe remove check_duplicate_indicator and check automatically?
        check_duplicate_indicator: bool = False,
        component: _score.Component | None = None,
        context: str | None = None,
        deactivate: bool = False,
        direction: _enums.Vertical | None = None,
        indicator: typing.Any | None = None,
        synthetic_offset: _duration.Offset | None = None,
        tag: _tag.Tag = _tag.Tag(),
    ) -> None:
        assert not isinstance(indicator, type(self)), repr(indicator)
        if annotation is not None:
            assert isinstance(annotation, str | enum.Enum), repr(annotation)
        self._annotation = annotation
        if component is not None:
            assert isinstance(component, _score.Component), repr(component)
        self._component = component
        deactivate = bool(deactivate)
        if context is not None:
            assert isinstance(context, str), repr(context)
        self._context = context
        if deactivate is not None:
            deactivate = bool(deactivate)
        self._deactivate = deactivate
        self._direction = direction
        self._effective_context = None
        self._indicator = indicator
        self._synthetic_offset: _duration.Offset | None
        if synthetic_offset is not None:
            assert isinstance(synthetic_offset, _duration.Offset), repr(
                synthetic_offset
            )
            self._synthetic_offset = _duration.Offset(synthetic_offset)
        else:
            self._synthetic_offset = synthetic_offset
        if tag is not None:
            assert isinstance(tag, str | _tag.Tag)
        assert isinstance(tag, _tag.Tag), repr(tag)
        self._tag = tag
        if component is not None:
            self._bind_component(
                component, check_duplicate_indicator=check_duplicate_indicator
            )

    def __copy__(self, *arguments) -> "Wrapper":
        r"""
        Copies wrapper.

        Preserves annotation flag:

        ..  container:: example

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> abjad.annotate(old_staff[0], "bow_direction", abjad.DOWN)
            >>> string = abjad.lilypond(old_staff)
            >>> print(string)
            \new Staff {
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> abjad.get.annotation(leaf, "bow_direction")
            <Vertical.DOWN: -1>

            >>> new_staff = abjad.mutate.copy(old_staff)
            >>> string = abjad.lilypond(new_staff)
            >>> print(string)
            \new Staff {
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> abjad.get.annotation(leaf, "bow_direction")
            <Vertical.DOWN: -1>

        ..  container:: example

            Preserves tag:

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> clef = abjad.Clef("alto")
            >>> abjad.attach(clef, old_staff[0], tag=abjad.Tag("RED:M1"))
            >>> string = abjad.lilypond(old_staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! M1
                %! RED
                \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> abjad.get.wrapper(leaf)
            Wrapper(annotation=None, context='Staff', deactivate=False, direction=None, indicator=Clef(name='alto', hide=False), synthetic_offset=None, tag=Tag(string='RED:M1'))

            >>> new_staff = abjad.mutate.copy(old_staff)
            >>> string = abjad.lilypond(new_staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! M1
                %! RED
                \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> abjad.get.wrapper(leaf)
            Wrapper(annotation=None, context='Staff', deactivate=False, direction=None, indicator=Clef(name='alto', hide=False), synthetic_offset=None, tag=Tag(string='RED:M1'))

        ..  container:: example

            Preserves deactivate flag:

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> abjad.attach(
            ...     abjad.Clef("alto"),
            ...     old_staff[0],
            ...     deactivate=True,
            ...     tag=abjad.Tag("RED:M1"),
            ... )
            >>> string = abjad.lilypond(old_staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! M1
                %! RED
                %@% \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> abjad.get.wrapper(leaf)
            Wrapper(annotation=None, context='Staff', deactivate=True, direction=None, indicator=Clef(name='alto', hide=False), synthetic_offset=None, tag=Tag(string='RED:M1'))

            >>> new_staff = abjad.mutate.copy(old_staff)
            >>> string = abjad.lilypond(new_staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! M1
                %! RED
                %@% \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> abjad.get.wrapper(leaf)
            Wrapper(annotation=None, context='Staff', deactivate=True, direction=None, indicator=Clef(name='alto', hide=False), synthetic_offset=None, tag=Tag(string='RED:M1'))

        Copies all properties except component.

        Copy operations must supply component after wrapper copy.
        """
        new = type(self)(
            annotation=self.annotation,
            component=None,
            context=self.context,
            deactivate=self.deactivate,
            direction=self.direction,
            indicator=copy.copy(self.indicator),
            synthetic_offset=self.synthetic_offset,
            tag=self.tag,
        )
        return new

    def __eq__(self, argument) -> bool:
        """
        Is true when self equals ``argument``.
        """
        if not isinstance(argument, Wrapper):
            return False
        if self.annotation != argument.annotation:
            return False
        if self.component != argument.component:
            return False
        if self.context != argument.context:
            return False
        if self.deactivate != argument.deactivate:
            return False
        if self.indicator != argument.indicator:
            return False
        if self.synthetic_offset != argument.synthetic_offset:
            return False
        if self.tag != argument.tag:
            return False
        return True

    def __hash__(self) -> int:
        """
        Hashes wrapper.
        """
        return hash(self.__class__.__name__ + str(self))

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        parameters = f"""
            annotation={self.annotation!r},
            context={self.context!r},
            deactivate={self.deactivate!r},
            direction={self.direction!r},
            indicator={self.indicator!r},
            synthetic_offset={self.synthetic_offset!r},
            tag={self.tag!r}
        """
        parameters = " ".join(parameters.split())
        return f"{type(self).__name__}({parameters})"

    def _bind_component(self, component, check_duplicate_indicator=False):
        if isinstance(self.indicator, _tweaks.Bundle):
            indicator = self.indicator.indicator
        else:
            indicator = self.indicator
        if getattr(indicator, "context", None) is not None:
            if check_duplicate_indicator is True:
                self._warn_duplicate_indicator(component)
            self._unbind_component()
            self._component = component
            self._update_effective_context()
            if getattr(indicator, "mutates_offsets_in_seconds", False):
                self._component._update_later(offsets_in_seconds=True)
        component._wrappers.append(self)

    def _bind_effective_context(self, correct_effective_context):
        self._unbind_effective_context()
        if correct_effective_context is not None:
            if self not in correct_effective_context._dependent_wrappers:
                correct_effective_context._dependent_wrappers.append(self)
        self._effective_context = correct_effective_context
        self._update_effective_context()
        if isinstance(self.indicator, _tweaks.Bundle):
            indicator = self.indicator.indicator
        else:
            indicator = self.indicator
        if correct_effective_context is not None:
            if getattr(indicator, "mutates_offsets_in_seconds", False):
                correct_effective_context._update_later(offsets_in_seconds=True)

    def _detach(self):
        self._unbind_component()
        self._unbind_effective_context()
        return self

    @staticmethod
    def _find_correct_effective_context(component, context):
        if context is None:
            return None
        abjad = importlib.import_module("abjad")
        context = getattr(abjad, context, context)
        candidate = None
        parentage = component._get_parentage()
        if isinstance(context, type):
            for component in parentage:
                if not hasattr(component, "_lilypond_type"):
                    continue
                if isinstance(component, context):
                    candidate = component
                    break
        elif isinstance(context, str):
            for component in parentage:
                if not hasattr(component, "_lilypond_type"):
                    continue
                if component.name == context or component.lilypond_type == context:
                    candidate = component
                    break
        else:
            raise TypeError("must be context or string: {context!r}.")
        if candidate.__class__.__name__ == "Voice":
            for component in reversed(parentage):
                if not component.__class__.__name__ == "Voice":
                    continue
                if component.name == candidate.name:
                    candidate = component
                    break
        return candidate

    def _get_effective_context(self):
        if self.component is not None:
            _updatelib._update_now(self.component, indicators=True)
        return self._effective_context

    def _unbind_component(self):
        if self._component is not None and id(self) in [
            id(_) for _ in self._component._wrappers
        ]:
            self._component._wrappers.remove(self)
        self._component = None

    def _unbind_effective_context(self):
        if (
            self._effective_context is not None
            and self in self._effective_context._dependent_wrappers
        ):
            self._effective_context._dependent_wrappers.remove(self)
        self._effective_context = None

    def _update_effective_context(self):
        correct_effective_context = self._find_correct_effective_context(
            self.component, self.context
        )
        if self._effective_context is not correct_effective_context:
            self._bind_effective_context(correct_effective_context)
        if correct_effective_context is not None:
            if self not in correct_effective_context._dependent_wrappers:
                correct_effective_context._dependent_wrappers.append(self)

    def _warn_duplicate_indicator(self, component):
        if self.deactivate is True:
            return
        if isinstance(self.indicator, _tweaks.Bundle):
            indicator = self.indicator.indicator
        else:
            indicator = self.indicator
        prototype = type(indicator)
        command = getattr(indicator, "command", None)
        wrapper = _getlib._get_effective(
            component,
            prototype,
            attributes={"command": command},
            unwrap=False,
        )
        wrapper_site = None
        if wrapper is not None:
            wrapper_site = getattr(wrapper.unbundle_indicator(), "site", None)
        my_site = getattr(indicator, "site", None)
        if (
            wrapper is None
            or wrapper.context is None
            or wrapper.deactivate is True
            or wrapper.start_offset != self.start_offset
            or wrapper_site != my_site
        ):
            return
        my_leak = getattr(indicator, "leak", None)
        if getattr(wrapper.unbundle_indicator(), "leak", None) != my_leak:
            return
        context = None
        for parent in component._get_parentage():
            if hasattr(parent, "_lilypond_type"):
                context = parent
                break
        wrapper_context = None
        for parent in wrapper.component._get_parentage():
            if hasattr(parent, "_lilypond_type"):
                wrapper_context = parent
                break
        if wrapper.unbundle_indicator() == indicator and context is not wrapper_context:
            return
        message = f"\n\nCan not attach ...\n\n{repr(self)}\n\n..."
        message += f" to {repr(component)}"
        message += f" in {getattr(context, 'name', None)} because ..."
        message += f"\n\n{repr(wrapper)}\n\n"
        message += "... is already attached"
        if component is wrapper.component:
            message += " to the same leaf."
        else:
            message += f" to {repr(wrapper.component)}"
            message += f" in {wrapper_context.name}."
        message += "\n"
        raise _exceptions.PersistentIndicatorError(message)

    @property
    def annotation(self) -> str | enum.Enum | None:
        """
        Gets wrapper annotation.

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> articulation = abjad.Articulation("accent")
            >>> abjad.attach(articulation, note, direction=abjad.UP)
            >>> wrapper = abjad.get.wrapper(note)
            >>> wrapper.annotation is None
            True

            >>> note = abjad.Note("c'4")
            >>> articulation = abjad.Articulation("accent")
            >>> abjad.annotate(note, "foo", articulation)
            >>> abjad.get.annotation(note, "foo")
            Articulation(name='accent')

        """
        return self._annotation

    @property
    def component(self):
        """
        Gets start component.

        Returns component or none.
        """
        return self._component

    @property
    def context(self) -> str | None:
        """
        Gets context (name).
        """
        return self._context

    @property
    def deactivate(self) -> bool:
        """
        Is true when wrapper deactivates tag.
        """
        assert self._deactivate in (True, False, None)
        return self._deactivate

    @deactivate.setter
    def deactivate(self, argument):
        assert argument in (True, False, None)
        self._deactivate = argument

    @property
    def direction(self):
        """
        Gets direction of indicator.
        """
        return self._direction

    @property
    def indicator(self) -> typing.Any:
        """
        Gets indicator.
        """
        return self._indicator

    @property
    def leaked_start_offset(self) -> _duration.Offset:
        r"""
        Gets start offset and checks to see whether indicator leaks to the right.

        This is either the wrapper's synthetic offset (if set); or the START offset of
        the wrapper's component (if indicator DOES NOT leak); or else the STOP offset of
        the wrapper's component (if indicator DOES leak).

        ..  container:: example

            Start- and stop-text-spans attach to the same leaf. But stop-text-span leaks
            to the right:

            >>> voice = abjad.Voice("c'2 d'2")
            >>> start_text_span = abjad.StartTextSpan()
            >>> abjad.attach(start_text_span, voice[0])
            >>> stop_text_span = abjad.StopTextSpan(leak=True)
            >>> abjad.attach(stop_text_span, voice[0])
            >>> abjad.show(voice) # doctest: +SKIP

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'2
                \startTextSpan
                <> \stopTextSpan
                d'2
            }

            Start offset and leaked start offset are the same for start-text-span:

            >>> wrapper = abjad.get.wrapper(voice[0], abjad.StartTextSpan)
            >>> wrapper.start_offset, wrapper.leaked_start_offset
            (Offset((0, 1)), Offset((0, 1)))

            Start offset and leaked start offset differ for stop-text-span:

            >>> wrapper = abjad.get.wrapper(voice[0], abjad.StopTextSpan)
            >>> wrapper.start_offset, wrapper.leaked_start_offset
            (Offset((0, 1)), Offset((1, 2)))

        """
        if self._synthetic_offset is not None:
            return self._synthetic_offset
        if isinstance(self.indicator, _tweaks.Bundle):
            indicator = self.indicator.indicator
        else:
            indicator = self.indicator
        assert isinstance(self.component, _score.Component)
        if not getattr(indicator, "leak", False):
            return self.component._get_timespan().start_offset
        else:
            return self.component._get_timespan().stop_offset

    @property
    def site_adjusted_start_offset(self) -> _duration.Offset:
        r"""
        Gets site-adjusted start offset.

        ..  container:: example

            >>> staff = abjad.Staff("c'4")
            >>> abjad.attach(abjad.Ottava(-1, site="before"), staff[0])
            >>> abjad.attach(abjad.Ottava(0, site="after"), staff[0])
            >>> for wrapper in abjad.get.wrappers(staff[0], abjad.Ottava):
            ...     wrapper.indicator, wrapper.site_adjusted_start_offset
            (Ottava(n=-1, site='before'), Offset((0, 1)))
            (Ottava(n=0, site='after'), Offset((1, 4)))

        Indicators with site equal to ``absolute_after``, ``after`` or ``closing``
        give a site-adjusted start offset equal to the stop offset of the wrapper's
        component.

        Indicators with any other site give a site-adjusted start offset equal to the
        start offset of the wrapper's component. This is the usual case, and means
        that site-adjusted start offset equals vanilla start offset.

        But if ``synthetic_offset`` is set then ``synthetic_offset`` is returned
        directly without examining the format site at all.
        """
        if self.synthetic_offset is not None:
            return self.synthetic_offset
        site = getattr(self.unbundle_indicator(), "site", "before")
        if site in ("absolute_after", "after", "closing"):
            return self.component._get_timespan().stop_offset
        else:
            return self.component._get_timespan().start_offset

    @property
    def start_offset(self) -> _duration.Offset:
        """
        Gets start offset.

        This is either the wrapper's synthetic offset or the start offset of the
        wrapper's component.
        """
        if self.synthetic_offset is not None:
            return self.synthetic_offset
        assert isinstance(self.component, _score.Component)
        return self.component._get_timespan().start_offset

    @property
    def synthetic_offset(self) -> _duration.Offset | None:
        """
        Gets synthetic offset.
        """
        return self._synthetic_offset

    @property
    def tag(self) -> _tag.Tag:
        """
        Gets and sets tag.
        """
        assert isinstance(self._tag, _tag.Tag), repr(self._tag)
        return self._tag

    @tag.setter
    def tag(self, argument):
        if not isinstance(argument, _tag.Tag):
            raise Exception(f"must be tag: {argument!r}.")
        self._tag = argument

    def bundled(self):
        """
        Is true when indicator is bundled.
        """
        return isinstance(self._indicator, _tweaks.Bundle)

    def get_item(self):
        """
        Gets indicator or bundled indicator.
        """
        return self._indicator

    def unbundle_indicator(self):
        """
        Gets unbundled indicator.
        """
        if isinstance(self.indicator, _tweaks.Bundle):
            return self.indicator.indicator
        return self.indicator


def _unsafe_attach(
    attachable,
    target,
    *,
    check_duplicate_indicator: bool = False,
    context: str | None = None,
    deactivate: bool = False,
    direction: _enums.Vertical | None = None,
    do_not_test: bool = False,
    synthetic_offset: _duration.Offset | None = None,
    tag: _tag.Tag | None = None,
    wrapper: bool = False,
) -> Wrapper | None:
    if isinstance(attachable, _tweaks.Bundle):
        nonbundle_attachable = attachable.indicator
    else:
        nonbundle_attachable = attachable
    assert not isinstance(nonbundle_attachable, _tweaks.Bundle)

    if isinstance(nonbundle_attachable, _tag.Tag):
        message = "use the tag=None keyword instead of attach():\n"
        message += f"   {repr(attachable)}"
        raise Exception(message)

    if tag is not None and not isinstance(tag, _tag.Tag):
        raise Exception(f"must be be tag: {repr(tag)}")

    assert nonbundle_attachable is not None, repr(nonbundle_attachable)
    assert isinstance(target, _score.Component), repr(target)

    grace_prototype = (_score.AfterGraceContainer, _score.BeforeGraceContainer)
    if context is not None and isinstance(nonbundle_attachable, grace_prototype):
        raise Exception(f"set context only for indicators, not {attachable!r}.")

    if deactivate is True and tag is None:
        raise Exception("tag must exist when deactivate is true.")

    if hasattr(nonbundle_attachable, "_attachment_test_all") and not do_not_test:
        result = nonbundle_attachable._attachment_test_all(target)
        if result is not True:
            assert isinstance(result, list), repr(result)
            result = ["  " + _ for _ in result]
            message = f"{nonbundle_attachable!r}._attachment_test_all():"
            result.insert(0, message)
            message = "\n".join(result)
            raise Exception(message)

    if isinstance(nonbundle_attachable, grace_prototype):
        if not isinstance(target, _score.Leaf):
            raise Exception("grace containers attach to single leaf only.")
        nonbundle_attachable._attach(target)
        return None

    if isinstance(target, _score.Container):
        acceptable = False
        if isinstance(
            nonbundle_attachable, dict | str | enum.Enum | _tag.Tag | Wrapper
        ):
            acceptable = True
        if getattr(nonbundle_attachable, "can_attach_to_containers", False):
            acceptable = True
        if not acceptable:
            message = f"can not attach {attachable!r} to containers: {target!r}"
            raise Exception(message)
    elif not isinstance(target, _score.Leaf):
        message = f"indicator {attachable!r} must attach to leaf, not {target!r}."
        raise Exception(message)

    component = target
    assert isinstance(component, _score.Component), repr(component)

    annotation = None
    if isinstance(attachable, Wrapper):
        annotation = attachable.annotation
        context = context or attachable.context
        deactivate = deactivate or attachable.deactivate
        synthetic_offset = synthetic_offset or attachable.synthetic_offset
        tag = tag or attachable.tag
        attachable._detach()
        attachable = attachable.indicator

    if hasattr(nonbundle_attachable, "context"):
        context = context or nonbundle_attachable.context

    if tag is None:
        tag = _tag.Tag()

    wrapper_ = Wrapper(
        annotation=annotation,
        check_duplicate_indicator=check_duplicate_indicator,
        component=component,
        context=context,
        deactivate=deactivate,
        direction=direction,
        indicator=attachable,
        synthetic_offset=synthetic_offset,
        tag=tag,
    )

    if wrapper is True:
        return wrapper_
    else:
        return None


def annotate(component, annotation, indicator) -> None:
    r"""
    Annotates ``component`` with ``indicator``.

    Annotates first note in staff:

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.annotate(staff[0], "bow_direction", abjad.DOWN)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        >>> abjad.get.annotation(staff[0], "bow_direction")
        <Vertical.DOWN: -1>

        >>> abjad.get.annotation(staff[0], "bow_fraction") is None
        True

        >>> abjad.get.annotation(staff[0], "bow_fraction", 99)
        99

    """
    if isinstance(annotation, _tag.Tag):
        message = "use the tag=None keyword instead of annotate():\n"
        message += f"   {repr(annotation)}"
        raise Exception(message)
    assert isinstance(annotation, str | enum.Enum), repr(annotation)
    Wrapper(annotation=annotation, component=component, indicator=indicator)


@typing.overload
def attach(
    attachable,
    target,
    *,
    check_duplicate_indicator: bool = False,
    context: str | None = None,
    deactivate: bool = False,
    direction: _enums.Vertical | None = None,
    do_not_test: bool = False,
    synthetic_offset: _duration.Offset | None = None,
    tag: _tag.Tag | None = None,
    wrapper: typing.Literal[True] = True,
) -> Wrapper:
    pass


@typing.overload
def attach(
    attachable,
    target,
    *,
    check_duplicate_indicator: bool = False,
    context: str | None = None,
    deactivate: bool = False,
    direction: _enums.Vertical | None = None,
    do_not_test: bool = False,
    synthetic_offset: _duration.Offset | None = None,
    tag: _tag.Tag | None = None,
    wrapper: typing.Literal[False],
) -> None:
    pass


def attach(
    attachable,
    target,
    *,
    check_duplicate_indicator: bool = False,
    context: str | None = None,
    deactivate: bool = False,
    direction: _enums.Vertical | None = None,
    do_not_test: bool = False,
    synthetic_offset: _duration.Offset | None = None,
    tag: _tag.Tag | None = None,
    wrapper: bool = False,
) -> Wrapper | None:
    r"""
    Attaches ``attachable`` to (leaf or container) ``target``.

    Acceptable types of ``attachable``:

    ::

        * indicator
        * abjad.Wrapper
        * abjad.BeforeGraceContainer
        * abjad.AfterGraceContainer

    The class of ``target`` is almost always ``abjad.Leaf``. A small number
    of indicators (like ``abjad.LilyPondLiteral``) can attach to both ``abjad.Leaf``
    and ``abjad.Container`` objects.

    Attaches clef to first note in staff:

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef("alto"), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Attaches accent to last note in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Articulation(">"), staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                - \accent
            }

    ..  container:: example

        Works with context names:

        >>> voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
        >>> staff = abjad.Staff([voice], name="MusicStaff")
        >>> abjad.attach(abjad.Clef("alto"), voice[0], context="MusicStaff")
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "MusicStaff"
            {
                \context Voice = "MusicVoice"
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    f'4
                }
            }

        >>> for leaf in abjad.select.leaves(staff):
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Note("c'4"), Clef(name='alto', hide=False))
        (Note("d'4"), Clef(name='alto', hide=False))
        (Note("e'4"), Clef(name='alto', hide=False))
        (Note("f'4"), Clef(name='alto', hide=False))

        Derives context from ``attachable.context`` context when ``context=None``.

    ..  container:: example

        If multiple contexted indicators are attached at the same offset then
        ``abjad.attach()`` raises ``abjad.PersistentIndicatorError`` if all
        indicators are active. But simultaneous contexted indicators are
        allowed if only one is active (and all others are inactive):

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef("treble"), staff[0])
        >>> abjad.attach(
        ...     abjad.Clef("alto"),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag=abjad.Tag("+PARTS"),
        ... )
        >>> abjad.attach(
        ...     abjad.Clef("tenor"),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag=abjad.Tag("+PARTS"),
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! +PARTS
                %@% \clef "alto"
                %! +PARTS
                %@% \clef "tenor"
                \clef "treble"
                c'4
                d'4
                e'4
                f'4
            }

        Active indicator is always effective when competing inactive indicators are
        present:

        >>> for note in staff:
        ...     clef = abjad.get.effective(staff[0], abjad.Clef)
        ...     note, clef
        ...
        (Note("c'4"), Clef(name='treble', hide=False))
        (Note("d'4"), Clef(name='treble', hide=False))
        (Note("e'4"), Clef(name='treble', hide=False))
        (Note("f'4"), Clef(name='treble', hide=False))

        But a lone inactivate indicator is effective when no active indicator is present:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(
        ...     abjad.Clef("alto"),
        ...     staff[0],
        ...     deactivate=True,
        ...     tag=abjad.Tag("+PARTS"),
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                %! +PARTS
                %@% \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

        >>> for note in staff:
        ...     clef = abjad.get.effective(staff[0], abjad.Clef)
        ...     note, clef
        ...
        (Note("c'4"), Clef(name='alto', hide=False))
        (Note("d'4"), Clef(name='alto', hide=False))
        (Note("e'4"), Clef(name='alto', hide=False))
        (Note("f'4"), Clef(name='alto', hide=False))

        Note that ``tag`` must be an ``abjad.Tag`` when ``deactivate=True``.

    Return type depends on value of ``wrapper``:

        * returns ``abjad.Wrapper`` when ``wrapper`` is true
        * returns none when ``wrapper`` is not true

    """
    if isinstance(attachable, _tweaks.Bundle):
        nonbundle_attachable = attachable.indicator
    else:
        nonbundle_attachable = attachable
    assert not isinstance(nonbundle_attachable, _tweaks.Bundle)
    assert nonbundle_attachable is not None, repr(nonbundle_attachable)
    assert isinstance(target, _score.Component), repr(target)
    _before_attach(nonbundle_attachable, context, deactivate, target)
    result = _unsafe_attach(
        attachable,
        target,
        check_duplicate_indicator=check_duplicate_indicator,
        context=context,
        deactivate=deactivate,
        direction=direction,
        do_not_test=do_not_test,
        synthetic_offset=synthetic_offset,
        tag=tag,
        wrapper=wrapper,
    )
    return result


def detach(argument, target=None, by_id=False):
    r"""
    Detaches indicators-equal-to-``argument`` from ``target``.

    Set ``by_id`` to true to detach exact ``argument`` from ``target`` (rather than
    detaching all indicators-equal-to-``argument``).

    ..  container:: example

        Detaches articulations from first note in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Articulation(">"), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \accent
                d'4
                e'4
                f'4
            }

        >>> abjad.detach(abjad.Articulation, staff[0])
        (Articulation(name='>'),)

        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

    ..  container:: example

        The use of ``by_id`` is motivated by the following.

        Consider the three document-specifier markups below:

        >>> markup_1 = abjad.Markup(r"\markup tutti")
        >>> markup_2 = abjad.Markup(r"\markup { with the others }")
        >>> markup_3 = abjad.Markup(r"\markup { with the others }")

        Markups two and three compare equal:

        >>> markup_2 == markup_3
        True

        But document-tagging like this makes sense for score and two diferent parts:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(markup_1, staff[0], direction=abjad.UP, tag=abjad.Tag(string="+SCORE"))
        >>> abjad.attach(
        ...     markup_2,
        ...     staff[0],
        ...     deactivate=True,
        ...     direction=abjad.UP,
        ...     tag=abjad.Tag("+PARTS_VIOLIN_1"),
        ... )
        >>> abjad.attach(
        ...     markup_3,
        ...     staff[0],
        ...     deactivate=True,
        ...     direction=abjad.UP,
        ...     tag=abjad.Tag("+PARTS_VIOLIN_2"),
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! +SCORE
            ^ \markup tutti
            %! +PARTS_VIOLIN_1
            %@% ^ \markup { with the others }
            %! +PARTS_VIOLIN_2
            %@% ^ \markup { with the others }
            d'4
            e'4
            f'4
        }

        The question is then how to detach just one of the two markups that compare equal
        to each other?

        Passing in one of the markup objects directory doesn't work. This is because
        detach tests for equality to input argument:

        >>> markups = abjad.detach(markup_2, staff[0])
        >>> for markup in markups:
        ...     markup
        Markup(string='\\markup { with the others }')
        Markup(string='\\markup { with the others }')

        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! +SCORE
            ^ \markup tutti
            d'4
            e'4
            f'4
        }

        We start again:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(markup_1, staff[0], direction=abjad.UP, tag=abjad.Tag(string="+SCORE"))
        >>> abjad.attach(
        ...     markup_2,
        ...     staff[0],
        ...     deactivate=True,
        ...     direction=abjad.UP,
        ...     tag=abjad.Tag("+PARTS_VIOLIN_1"),
        ... )
        >>> abjad.attach(
        ...     markup_3,
        ...     staff[0],
        ...     deactivate=True,
        ...     direction=abjad.UP,
        ...     tag=abjad.Tag("+PARTS_VIOLIN_2"),
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! +SCORE
            ^ \markup tutti
            %! +PARTS_VIOLIN_1
            %@% ^ \markup { with the others }
            %! +PARTS_VIOLIN_2
            %@% ^ \markup { with the others }
            d'4
            e'4
            f'4
        }

        This time we set ``by_id`` to true. Now detach checks the exact id of its input
        argument (rather than just testing for equality). This gives us what we want:

        >>> markups = abjad.detach(markup_2, staff[0], by_id=True)
        >>> for markup in markups:
        ...     markup
        Markup(string='\\markup { with the others }')

        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! +SCORE
            ^ \markup tutti
            %! +PARTS_VIOLIN_2
            %@% ^ \markup { with the others }
            d'4
            e'4
            f'4
        }

    ..  container:: example

        REGRESSION. Attach-detach-attach pattern works correctly when detaching wrappers:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(abjad.Clef("alto"), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "alto"
                c'4
                d'4
                e'4
                f'4
            }

        >>> wrapper = abjad.get.wrappers(staff[0])[0]
        >>> abjad.detach(wrapper, wrapper.component)
        (Wrapper(annotation=None, context='Staff', deactivate=False, direction=None, indicator=Clef(name='alto', hide=False), synthetic_offset=None, tag=Tag(string='')),)

        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        >>> abjad.attach(abjad.Clef("tenor"), staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "tenor"
                c'4
                d'4
                e'4
                f'4
            }

    Returns tuple of zero or more detached items.
    """
    assert target is not None
    after_grace_container = None
    before_grace_container = None
    if isinstance(argument, type):
        if "AfterGraceContainer" in argument.__name__:
            after_grace_container = target._after_grace_container
        elif "BeforeGraceContainer" in argument.__name__:
            before_grace_container = target._before_grace_container
        else:
            assert hasattr(target, "_wrappers")
            result = []
            for wrapper in target._wrappers[:]:
                if isinstance(wrapper, argument):
                    target._wrappers.remove(wrapper)
                    result.append(wrapper)
                elif isinstance(wrapper.unbundle_indicator(), argument):
                    wrapper._detach()
                    result.append(wrapper.get_item())
            result = tuple(result)
            return result
    else:
        if "AfterGraceContainer" in argument.__class__.__name__:
            after_grace_container = target._after_grace_container
        elif "BeforeGraceContainer" in argument.__class__.__name__:
            before_grace_container = target._before_grace_container
        else:
            assert hasattr(target, "_wrappers")
            result = []
            for wrapper in target._wrappers[:]:
                if wrapper is argument:
                    wrapper._detach()
                    result.append(wrapper)
                elif wrapper.unbundle_indicator() == argument:
                    if by_id is True and id(argument) != id(
                        wrapper.unbundle_indicator()
                    ):
                        pass
                    else:
                        wrapper._detach()
                        result.append(wrapper.get_item())
            result = tuple(result)
            return result
    items = []
    if after_grace_container is not None:
        items.append(after_grace_container)
    if before_grace_container is not None:
        items.append(before_grace_container)
    if by_id is True:
        items = [_ for _ in items if id(_) == id(argument)]
    for item in items:
        item._detach()
    items = tuple(items)
    return items



import copy
import dataclasses
import typing

from . import _indentlib
from . import contributions as _contributions
from . import enums as _enums
from . import lyenv as _lyenv
from . import string as _string


def _format_scheme_value(value):
    if value is True:
        result = "##t"
    elif value is False:
        result = "##f"
    elif isinstance(value, int | float):
        result = str(value)
    elif value is None:
        result = "##f"
    elif isinstance(value, _enums.Horizontal | _enums.Vertical):
        result = rf"#{value.name.lower()}"
    elif isinstance(value, tuple) and len(value) == 2:
        result = f"#'({value[0]} . {value[1]})"
    else:
        assert isinstance(value, str), repr(value)
        result = str(value)
    return result


def _format_lilypond_attribute(attribute):
    assert isinstance(attribute, str), repr(attribute)
    attribute = attribute.replace("__", ".")
    result = attribute.replace("_", "-")
    return result


def _make_lilypond_tweak_string(attribute, value, *, directed=True, grob=None) -> str:
    if grob is not None:
        grob = _string.to_upper_camel_case(grob)
        grob += "."
    else:
        grob = ""
    attribute = _format_lilypond_attribute(attribute)
    value = _format_scheme_value(value)
    string = rf"\tweak {grob}{attribute} {value}"
    if directed:
        string = "- " + string
    return string


@dataclasses.dataclass(order=True, slots=True, unsafe_hash=True)
class LilyPondOverride:
    r"""
    LilyPond grob override.

    ..  container:: example

        >>> override = abjad.LilyPondOverride(
        ...    lilypond_type="Staff",
        ...    grob_name="TextSpanner",
        ...    once=True,
        ...    property_path=(
        ...        "bound-details",
        ...        "left",
        ...        "text",
        ...    ),
        ...    value=r"\markup \bold { over pressure }",
        ... )
        >>> print(override.override_string)
        \once \override Staff.TextSpanner.bound-details.left.text = \markup \bold { over pressure }

        >>> override = abjad.LilyPondOverride(
        ...    lilypond_type="Staff",
        ...    grob_name="TextSpanner",
        ...    once=True,
        ...    property_path="bound_details__left__text",
        ...    value=r"\markup \bold { over pressure }",
        ... )
        >>> print(override.override_string)
        \once \override Staff.TextSpanner.bound-details.left.text = \markup \bold { over pressure }

    """

    lilypond_type: str | None = None
    grob_name: str = "NoteHead"
    once: bool = False
    is_revert: bool = False
    property_path: str | typing.Sequence[str] = "color"
    value: bool | int | float | str = "#red"

    format_leaf_children: typing.ClassVar[bool] = False

    def __post_init__(self):
        if self.lilypond_type is not None:
            assert isinstance(self.lilypond_type, str)
        assert isinstance(self.grob_name, str)
        assert isinstance(self.once, bool), repr(self.once)
        assert isinstance(self.is_revert, bool)
        if isinstance(self.property_path, str):
            property_path_ = (self.property_path,)
        else:
            property_path_ = tuple(self.property_path)
        assert all(isinstance(_, str) for _ in property_path_)
        self.property_path = property_path_
        prototype = (
            bool | int | float | str | _enums.Horizontal | _enums.Vertical | tuple
        )
        assert isinstance(self.value, prototype), repr(self.value)

    def _get_contributions(self, component=None):
        contributions = _contributions.ContributionsBySite()
        if not self.once:
            revert_format = "\n".join(self.revert_format_pieces)
            contributions.grob_reverts.append(revert_format)
        if not self.is_revert:
            override_format = "\n".join(self.override_format_pieces)
            contributions.grob_overrides.append(override_format)
        return contributions

    def _override_property_path_string(self):
        parts = []
        if self.lilypond_type is not None:
            parts.append(self.lilypond_type)
        parts.append(self.grob_name)
        for part in self.property_path:
            part = _format_lilypond_attribute(part)
            parts.append(part)
        path = ".".join(parts)
        return path

    def _revert_property_path_string(self):
        parts = []
        if self.lilypond_type is not None:
            parts.append(self.lilypond_type)
        parts.append(self.grob_name)
        part = _format_lilypond_attribute(self.property_path[0])
        parts.append(part)
        path = ".".join(parts)
        return path

    @property
    def override_format_pieces(self) -> tuple[str, ...]:
        r"""
        Gets LilyPond grob override \override format pieces.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...    ),
            ...    value=r"\markup \bold { over pressure }",
            ... )
            >>> for line in override.override_format_pieces:
            ...     line
            ...
            '\\once \\override Staff.TextSpanner.bound-details.left.text = \\markup \\bold { over pressure }'

        """
        result = []
        if self.once:
            result.append(r"\once")
        result.append(r"\override")
        result.append(self._override_property_path_string())
        result.append("=")
        string = _format_scheme_value(self.value)
        value_pieces = string.split("\n")
        result.append(value_pieces[0])
        result[:] = [" ".join(result)]
        result.extend(value_pieces[1:])
        return tuple(result)

    @property
    def override_string(self) -> str:
        r"""
        Gets LilyPond grob override \override string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ... )
            >>> override.override_string
            "\\override Glissando.style = #'zigzag"

        """
        return "\n".join(self.override_format_pieces)

    @property
    def revert_format_pieces(self) -> tuple[str, ...]:
        r"""
        Gets LilyPond grob override \revert format pieces.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ... )
            >>> override.revert_format_pieces
            ('\\revert Glissando.style',)

        """
        result = rf"\revert {self._revert_property_path_string()}"
        return (result,)

    @property
    def revert_string(self) -> str:
        r"""
        Gets LilyPond grob override \revert string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ... )
            >>> override.revert_string
            '\\revert Glissando.style'

        """
        return "\n".join(self.revert_format_pieces)

    def tweak_string(self, directed=True, grob=False) -> str:
        r"""
        Gets LilyPond grob override \tweak string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ... )
            >>> override.tweak_string()
            "- \\tweak style #'zigzag"

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="RehearsalMark",
            ...     property_path="color",
            ...     value="#red",
            ... )
            >>> override.tweak_string(directed=False)
            '\\tweak color #red'

        """
        if directed:
            result = [r"- \tweak"]
        else:
            result = [r"\tweak"]
        if grob:
            property_path = [self.grob_name] + list(self.property_path)
        else:
            property_path = list(self.property_path)
        string = ".".join(property_path)
        result.append(string)
        string = _format_scheme_value(self.value)
        result.append(string)
        return " ".join(result)


@dataclasses.dataclass(order=True, slots=True, unsafe_hash=True)
class LilyPondSetting:
    r"""
    LilyPond context setting.

    ..  container:: example

        >>> context_setting = abjad.LilyPondSetting(
        ...    lilypond_type="Score",
        ...    context_property="autoBeaming",
        ...    value="##f",
        ... )

        >>> print("\n".join(context_setting.format_pieces))
        \set Score.autoBeaming = ##f

    """

    lilypond_type: str | None = None
    context_property: str = "autoBeaming"
    is_unset: bool = False
    value: typing.Any = False

    format_leaf_children: typing.ClassVar[bool] = False

    def __post_init__(self):
        if self.lilypond_type is not None:
            assert isinstance(self.lilypond_type, str)
        assert isinstance(self.context_property, str)
        assert isinstance(self.is_unset, bool)
        assert isinstance(self.value, bool | int | float | str), repr(self.value)

    def _get_contributions(self, component=None):
        contributions = _contributions.ContributionsBySite()
        string = "\n".join(self.format_pieces)
        contributions.context_settings.append(string)
        return contributions

    @property
    def format_pieces(self) -> tuple[str, ...]:
        r"""
        Gets LilyPond context setting ``\set`` or ``\unset`` format pieces.
        """
        result = []
        if not self.is_unset:
            result.append(r"\set")
        else:
            result.append(r"\unset")
        if self.lilypond_type is not None:
            string = f"{self.lilypond_type}.{self.context_property}"
            result.append(string)
        else:
            result.append(self.context_property)
        result.append("=")
        string = _format_scheme_value(self.value)
        value_pieces = string.split("\n")
        result.append(value_pieces[0])
        result[:] = [" ".join(result)]
        result.extend(value_pieces[1:])
        return tuple(result)


class Interface:
    """
    Base class from which grob, setting and tweak interfaces inherit.
    """

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is an interface with attribute pairs equal to those of
        this interface.

        ..  container:: example

            >>> note_1 = abjad.Note("c'4")
            >>> abjad.setting(note_1).Voice.auto_beaming = False
            >>> abjad.setting(note_1).Staff.tupletFullLength = True

            >>> note_2 = abjad.Note("c'4")
            >>> abjad.setting(note_2).Voice.auto_beaming = False
            >>> abjad.setting(note_2).Staff.tupletFullLength = True

            >>> note_3 = abjad.Note("c'4")
            >>> abjad.setting(note_3).Voice.auto_beaming = True

            >>> setting_1 = abjad.setting(note_1)
            >>> setting_2 = abjad.setting(note_2)
            >>> setting_3 = abjad.setting(note_3)

            >>> setting_1 == setting_1
            True
            >>> setting_1 == setting_2
            True
            >>> setting_1 == setting_3
            False

            >>> setting_2 == setting_1
            True
            >>> setting_2 == setting_2
            True
            >>> setting_2 == setting_3
            False

            >>> setting_3 == setting_1
            False
            >>> setting_3 == setting_2
            False
            >>> setting_3 == setting_3
            True

        ..  container:: example

            >>> note_1 = abjad.Note("c'4")
            >>> abjad.override(note_1).NoteHead.color = "#red"
            >>> abjad.override(note_1).Stem.color = "#red"

            >>> note_2 = abjad.Note("c'4")
            >>> abjad.override(note_2).NoteHead.color = "#red"
            >>> abjad.override(note_2).Stem.color = "#red"

            >>> note_3 = abjad.Note("c'4")
            >>> abjad.override(note_3).NoteHead.color = "#red"

            >>> override_1 = abjad.override(note_1)
            >>> override_2 = abjad.override(note_2)
            >>> override_3 = abjad.override(note_3)

            >>> override_1 == override_1
            True
            >>> override_1 == override_2
            True
            >>> override_1 == override_3
            False

            >>> override_2 == override_1
            True
            >>> override_2 == override_2
            True
            >>> override_2 == override_3
            False

            >>> override_3 == override_1
            False
            >>> override_3 == override_2
            False
            >>> override_3 == override_3
            True

        """
        if isinstance(argument, type(self)):
            attribute_pairs_1 = self._get_attribute_pairs()
            attribute_pairs_2 = argument._get_attribute_pairs()
            return attribute_pairs_1 == attribute_pairs_2
        return False

    def __getstate__(self) -> dict:
        """
        Gets object state.
        """
        return copy.deepcopy(vars(self))

    def __hash__(self) -> int:
        """
        Hashes interface.
        """
        return super().__hash__()

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        body_string = ""
        pairs = self._get_attribute_pairs()
        pairs = [str(_) for _ in pairs]
        body_string = ", ".join(pairs)
        return f"{type(self).__name__}({body_string})"

    def __setstate__(self, state) -> None:
        """
        Sets object state.
        """
        for key, value in state.items():
            self.__dict__[key] = value

    def _get_attribute_pairs(self):
        return list(sorted(vars(self).items()))


class OverrideInterface(Interface):
    """
    Override interface.

    ..  container:: example

        Override interfaces are created by the ``abjad.override()`` factory function:

        >>> note = abjad.Note("c'4")
        >>> abjad.override(note)
        OverrideInterface()

    """

    def __getattr__(self, name):
        r"""
        Gets Interface (or OverrideInterface) keyed to ``name``.

        ..  container:: example

            Somewhat confusingly, getting a grob name returns an interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).NoteHead
            Interface()

            While getting a context name returns an override interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).Staff
            OverrideInterface()

            Which can then be deferenced to get an interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).Staff.NoteHead
            Interface()

        Note that the dot-chained user syntax is unproblematic. But the class of each
        interface returned in the chain is likely to be surprising at first encounter.
        """
        camel_name = _string.to_upper_camel_case(name)
        if name.startswith("_"):
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError("{type_name!r} object has no attribute: {name!r}.")
        elif camel_name in _lyenv.contexts:
            try:
                return vars(self)["_" + name]
            except KeyError:
                context = OverrideInterface()
                vars(self)["_" + name] = context
                return context
        elif camel_name in _lyenv.grob_interfaces:
            try:
                return vars(self)[name]
            except KeyError:
                vars(self)[name] = Interface()
                return vars(self)[name]
        else:
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError(
                    f"{type_name!r} object has no attribute: {name!r}."
                )

    def __setattr__(self, attribute, value):
        """
        Sets attribute ``attribute`` of override interface to ``value``.
        """
        object.__setattr__(self, attribute, value)

    def _get_attribute_tuples(self):
        result = []
        for name, value in vars(self).items():
            if type(value) is Interface:
                grob, grob_proxy = name, value
                pairs = iter(vars(grob_proxy).items())
                for attribute, value in pairs:
                    triple = (grob, attribute, value)
                    result.append(triple)
            else:
                context, context_proxy = name.strip("_"), value
                for grob, grob_proxy in vars(context_proxy).items():
                    pairs = iter(vars(grob_proxy).items())
                    for attribute, value in pairs:
                        quadruple = (context, grob, attribute, value)
                        result.append(quadruple)
        result.sort()
        return tuple(result)

    def _list_contributions(self, contribution_type, once=False):
        assert contribution_type in ("override", "revert")
        result = []
        for attribute_tuple in self._get_attribute_tuples():
            if len(attribute_tuple) == 3:
                context = None
                grob = attribute_tuple[0]
                attribute = attribute_tuple[1]
                value = attribute_tuple[2]
            elif len(attribute_tuple) == 4:
                context = attribute_tuple[0]
                grob = attribute_tuple[1]
                attribute = attribute_tuple[2]
                value = attribute_tuple[3]
            else:
                raise ValueError(f"invalid attribute tuple: {attribute_tuple!r}.")
            if contribution_type == "override":
                override = LilyPondOverride(
                    lilypond_type=context,
                    grob_name=grob,
                    once=once,
                    is_revert=False,
                    property_path=attribute,
                    value=value,
                )
                override_string = override.override_string
                result.append(override_string)
            else:
                override = LilyPondOverride(
                    lilypond_type=context,
                    grob_name=grob,
                    is_revert=True,
                    property_path=attribute,
                )
                revert_string = override.revert_string
                result.append(revert_string)
        result.sort()
        return result


class SettingInterface(Interface):
    """
    LilyPond setting interface.

    ..  container:: example

        >>> note = abjad.Note("c'4")
        >>> abjad.setting(note)
        SettingInterface()

    """

    def __getattr__(self, name: str):
        r"""
        Gets arbitrary object keyed to ``name``.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> string = r'\markup "Vn. I"'
            >>> abjad.setting(staff).instrumentName = string
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                \with
                {
                    instrumentName = \markup "Vn. I"
                }
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }

        ..  container:: example

            Returns arbitrary object keyed to ``name``:

            >>> abjad.setting(staff).instrumentName
            '\\markup "Vn. I"'

        """
        camel_name = _string.to_upper_camel_case(name)
        if name.startswith("_"):
            try:
                return vars(self)[name]
            except KeyError:
                message = "{type(self).__name__!r} object has no attribute: {name!r}."
                raise AttributeError(message)
        elif camel_name in _lyenv.contexts:
            try:
                return vars(self)["_" + name]
            except KeyError:
                context = Interface()
                vars(self)["_" + name] = context
                return context
        else:
            try:
                return vars(self)[name]
            except KeyError:
                message = "{type(self).__name__!r} object has no attribute: {name!r}."
                raise AttributeError(message)

    def _format_in_with_block(self) -> list[str]:
        strings = []
        for key, value in vars(self).items():
            assert isinstance(key, str), repr(key)
            name = key.split("_")
            first = name[0:1]
            rest = name[1:]
            rest = [_.title() for _ in rest]
            name = first + rest
            string = "".join(name)
            value = _format_scheme_value(value)
            value_parts = value.split("\n")
            result = rf"{string!s} = {value_parts[0]!s}"
            pieces = [result]
            for part in value_parts[1:]:
                pieces.append(_indentlib.INDENT + part)
            string = "\n".join(pieces)
            strings.append(string)
        return strings

    def _format_inline(self, context=None) -> list[str]:
        result = []
        for name, value in vars(self).items():
            # if we've found a leaf context namespace
            if name.startswith("_"):
                for x, y in vars(value).items():
                    if not x.startswith("_"):
                        string = self._format_inline_helper(x, y, name)
                        result.append(string)
            # otherwise we've found a default leaf context setting
            else:
                # parse default context setting
                string = self._format_inline_helper(name, value)
                result.append(string)
        return result

    def _format_inline_helper(self, name, value, context=None):
        name = name.split("_")
        first = name[0:1]
        rest = name[1:]
        rest = [_.title() for _ in rest]
        name = first + rest
        name = "".join(name)
        value = _format_scheme_value(value)
        if context is not None:
            context_string = context[1:]
            context_string = context_string.split("_")
            context_string = [_.title() for _ in context_string]
            context_string = "".join(context_string)
            context_string += "."
        else:
            context_string = ""
        result = rf"\set {context_string}{name} = {value}"
        return result

    def _get_attribute_tuples(self):
        result = []
        for name, value in vars(self).items():
            if type(value) is Interface:
                prefixed_context_name = name
                lilypond_type = prefixed_context_name.strip("_")
                context_proxy = value
                attribute_pairs = context_proxy._get_attribute_pairs()
                for attribute_name, attribute_value in attribute_pairs:
                    triple = (lilypond_type, attribute_name, attribute_value)
                    result.append(triple)
            else:
                attribute_name, attribute_value = name, value
                result.append((attribute_name, attribute_value))
        result.sort()
        return result


def override(argument):
    r"""
    Makes LilyPond grob override interface.

    ..  container:: example

        Overrides staff symbol color:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.override(staff).StaffSymbol.color = "#red"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override StaffSymbol.color = #red
            }
            {
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Specify grob context like this:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.override(staff[0]).Staff.StaffSymbol.color = "#blue"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \once \override Staff.StaffSymbol.color = #blue
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Returns LilyPond grob override interface.

        >>> staff = abjad.Staff("c'4 e' d' f'")
        >>> abjad.override(staff)
        OverrideInterface()

    """
    if getattr(argument, "_overrides", None) is None:
        argument._overrides = OverrideInterface()
    return argument._overrides


def setting(argument):
    r"""
    Makes LilyPond setting name interface.

    ..  container:: example

        Sets instrument name:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> string = r'\markup "Vn. I"'
        >>> abjad.setting(staff).instrumentName = string
        >>> abjad.show(staff) # doctest: +SKIP


        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                instrumentName = \markup "Vn. I"
            }
            {
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Returns LilyPond setting name interface:

        >>> abjad.setting(staff)
        SettingInterface(('instrumentName', '\\markup "Vn. I"'))

    """
    if getattr(argument, "_lilypond_setting_name_manager", None) is None:
        argument._lilypond_setting_name_manager = SettingInterface()
    return argument._lilypond_setting_name_manager



#! /usr/bin/env python
import os

import setuptools


def get_abjad_version():
    version_file_path = os.path.join(os.path.dirname(__file__), "abjad", "_version.py")
    with open(version_file_path, "r") as file_pointer:
        file_contents_string = file_pointer.read()
    local_dict: dict = {}
    exec(file_contents_string, None, local_dict)
    __version__ = local_dict["__version__"]
    return __version__


with open("README.rst", "r") as file_pointer:
    long_description = file_pointer.read()


if __name__ == "__main__":
    version = get_abjad_version()
    setuptools.setup(
        author="Trevor Bača, Joséphine Wolf Oberholtzer",
        author_email="trevor.baca@gmail.com, josephine.wolf.oberholtzer@gmail.com",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: GNU General Public License (GPL)",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Topic :: Artistic Software",
        ],
        description="Abjad is a Python API for building LilyPond files.",
        extras_require={
            "nauert": ["abjad-ext-nauert>=3.21"],
            "rmakers": ["abjad-ext-rmakers>=3.21"],
            "dev": [
                "black>=25.1.0",
                "flake8>=7.1.1",
                "isort>=6.0.0",
                "mypy>=1.14.1",
                "pytest>=8.3.4",
                "pytest-cov>=6.0.0",
                "pytest-helpers-namespace>=2021.12.29",
                "setuptools>=75.8.0",
                "sphinx>=8.1.3",
                "sphinx-rtd-theme>=3.0.2",
            ],
        },
        include_package_data=True,
        install_requires=["ply>=3.11", "roman>=5.0", "uqbar>=0.7.4"],
        keywords="lilypond, music composition, music notation",
        license="MIT",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        name="abjad",
        packages=["abjad"],
        platforms="Any",
        python_requires=">=3.12",
        url="https://abjad.github.io",
        version=version,
    )



r"""
..  container:: example

    **CORNER CASE 1.**

    Before-grace-to-on-beat-grace works correctly when before-grace container
    is attached first:

    >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
    >>> container = abjad.BeforeGraceContainer("gs'16")
    >>> abjad.attach(container, music_voice[1])
    >>> obgc = abjad.on_beat_grace_container(
    ...     "a'8 b' c'' b'", music_voice[1:3], grace_leaf_duration=(1, 24)
    ... )
    >>> abjad.attach(abjad.Articulation(">"), obgc[0])
    >>> staff = abjad.Staff([music_voice])
    >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
    >>> abjad.show(lilypond_file) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            \context Voice = "MusicVoice"
            {
                c'4
                <<
                    \context Voice = "On_Beat_Grace_Container"
                    {
                        \set fontSize = #-3
                        \slash
                        \voiceOne
                        <
                            \tweak font-size 0
                            \tweak transparent ##t
                            d'
                            a'
                        >8 * 1/3
                        - \accent
                        [
                        (
                        b'8 * 1/3
                        c''8 * 1/3
                        b'8 * 1/3
                        )
                        ]
                    }
                    \context Voice = "MusicVoice"
                    {
                        \grace {
                            gs'16
                        }
                        \voiceTwo
                        d'4
                        e'4
                    }
                >>
                \oneVoice
                f'4
            }
        }

    Before-grace-to-on-beat-grace works correctly when on-beat grace container
    is attached first:

    >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
    >>> obgc = abjad.on_beat_grace_container(
    ...     "a'8 b' c'' b'", music_voice[1:3], grace_leaf_duration=(1, 24)
    ... )
    >>> abjad.attach(abjad.Articulation(">"), obgc[0])
    >>> container = abjad.BeforeGraceContainer("gs'16")
    >>> abjad.attach(container, music_voice[1][1][0])
    >>> staff = abjad.Staff([music_voice])
    >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
    >>> abjad.show(lilypond_file) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            \context Voice = "MusicVoice"
            {
                c'4
                <<
                    \context Voice = "On_Beat_Grace_Container"
                    {
                        \set fontSize = #-3
                        \slash
                        \voiceOne
                        <
                            \tweak font-size 0
                            \tweak transparent ##t
                            d'
                            a'
                        >8 * 1/3
                        - \accent
                        [
                        (
                        b'8 * 1/3
                        c''8 * 1/3
                        b'8 * 1/3
                        )
                        ]
                    }
                    \context Voice = "MusicVoice"
                    {
                        \grace {
                            gs'16
                        }
                        \voiceTwo
                        d'4
                        e'4
                    }
                >>
                \oneVoice
                f'4
            }
        }

..  container:: example

    **CORNER CASE 2.** After-grace-within-on-beat-grace works correctly:

    >>> music_voice = abjad.Voice("c'4 d' e' f'", name="MusicVoice")
    >>> container = abjad.AfterGraceContainer("cs'16")
    >>> abjad.attach(container, music_voice[1])
    >>> obgc = abjad.on_beat_grace_container(
    ...     "a'8 b' c'' b'", music_voice[1:3], grace_leaf_duration=(1, 24)
    ... )
    >>> abjad.attach(abjad.Articulation(">"), obgc[0])
    >>> staff = abjad.Staff([music_voice])
    >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
    >>> abjad.show(lilypond_file) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            \context Voice = "MusicVoice"
            {
                c'4
                <<
                    \context Voice = "On_Beat_Grace_Container"
                    {
                        \set fontSize = #-3
                        \slash
                        \voiceOne
                        <
                            \tweak font-size 0
                            \tweak transparent ##t
                            d'
                            a'
                        >8 * 1/3
                        - \accent
                        [
                        (
                        b'8 * 1/3
                        c''8 * 1/3
                        b'8 * 1/3
                        )
                        ]
                    }
                    \context Voice = "MusicVoice"
                    {
                        \voiceTwo
                        \afterGrace
                        d'4
                        {
                            cs'16
                        }
                        e'4
                    }
                >>
                \oneVoice
                f'4
            }
        }

..  container:: example

    **CORNER CASE 3.** After-grace-to-before-grace works correctly:

    >>> music_voice = abjad.Voice("b4 d' e' f'", name="MusicVoice")
    >>> container = abjad.AfterGraceContainer("c'16")
    >>> abjad.attach(container, music_voice[0])
    >>> container = abjad.BeforeGraceContainer("cs'16")
    >>> abjad.attach(container, music_voice[1])
    >>> staff = abjad.Staff([music_voice])
    >>> abjad.show(staff) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            \context Voice = "MusicVoice"
            {
                \afterGrace
                b4
                {
                    c'16
                }
                \grace {
                    cs'16
                }
                d'4
                e'4
                f'4
            }
        }

    >>> for component in abjad.iterate.components(staff):
    ...     timespan = abjad.get.timespan(component)
    ...     print(f"{repr(component):30} {repr(timespan)}")
    Staff("{ b4 d'4 e'4 f'4 }")    Timespan(Offset((0, 1)), Offset((1, 1)))
    Voice("b4 d'4 e'4 f'4", name='MusicVoice') Timespan(Offset((0, 1)), Offset((1, 1)))
    Note('b4')                     Timespan(Offset((0, 1)), Offset((1, 4)))
    AfterGraceContainer("c'16")    Timespan(Offset((1, 4), displacement=Duration(-1, 8)), Offset((1, 4), displacement=Duration(-1, 16)))
    Note("c'16")                   Timespan(Offset((1, 4), displacement=Duration(-1, 8)), Offset((1, 4), displacement=Duration(-1, 16)))
    BeforeGraceContainer("cs'16")  Timespan(Offset((1, 4), displacement=Duration(-1, 16)), Offset((1, 4)))
    Note("cs'16")                  Timespan(Offset((1, 4), displacement=Duration(-1, 16)), Offset((1, 4)))
    Note("d'4")                    Timespan(Offset((1, 4)), Offset((1, 2)))
    Note("e'4")                    Timespan(Offset((1, 2)), Offset((3, 4)))
    Note("f'4")                    Timespan(Offset((3, 4)), Offset((1, 1)))

..  container:: example

    **CORNER CASE 4.** After-grace-to-on-beat-grace works correctly:

    >>> music_voice = abjad.Voice("b4 d' e' f'", name="MusicVoice")
    >>> container = abjad.AfterGraceContainer("c'16")
    >>> abjad.attach(container, music_voice[0])
    >>> obgc = abjad.on_beat_grace_container(
    ...     "a'8 b' c'' b'", music_voice[1:3], grace_leaf_duration=(1, 24)
    ... )
    >>> abjad.attach(abjad.Articulation(">"), obgc[0])
    >>> staff = abjad.Staff([music_voice])
    >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', staff])
    >>> abjad.show(lilypond_file) # doctest: +SKIP

    ..  docs::

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            \context Voice = "MusicVoice"
            {
                \afterGrace
                b4
                {
                    c'16
                }
                <<
                    \context Voice = "On_Beat_Grace_Container"
                    {
                        \set fontSize = #-3
                        \slash
                        \voiceOne
                        <
                            \tweak font-size 0
                            \tweak transparent ##t
                            d'
                            a'
                        >8 * 1/3
                        - \accent
                        [
                        (
                        b'8 * 1/3
                        c''8 * 1/3
                        b'8 * 1/3
                        )
                        ]
                    }
                    \context Voice = "MusicVoice"
                    {
                        \voiceTwo
                        d'4
                        e'4
                    }
                >>
                \oneVoice
                f'4
            }
        }

    >>> for component in abjad.iterate.components(staff):
    ...     timespan = abjad.get.timespan(component)
    ...     print(f"{repr(component):30} {repr(timespan)}")
    Staff("{ b4 { { <d' a'>8 * 1/3 b'8 * 1/3 c''8 * 1/3 b'8 * 1/3 } { d'4 e'4 } } f'4 }") Timespan(Offset((0, 1)), Offset((1, 1)))
    Voice("b4 { { <d' a'>8 * 1/3 b'8 * 1/3 c''8 * 1/3 b'8 * 1/3 } { d'4 e'4 } } f'4", name='MusicVoice') Timespan(Offset((0, 1)), Offset((1, 1)))
    Note('b4')                     Timespan(Offset((0, 1)), Offset((1, 4)))
    AfterGraceContainer("c'16")    Timespan(Offset((1, 4), displacement=Duration(-1, 16)), Offset((1, 4)))
    Note("c'16")                   Timespan(Offset((1, 4), displacement=Duration(-1, 16)), Offset((1, 4)))
    Container("{ <d' a'>8 * 1/3 b'8 * 1/3 c''8 * 1/3 b'8 * 1/3 } { d'4 e'4 }") Timespan(Offset((1, 4)), Offset((3, 4)))
    OnBeatGraceContainer("<d' a'>8 * 1/3 b'8 * 1/3 c''8 * 1/3 b'8 * 1/3") Timespan(Offset((1, 4)), Offset((1, 4), displacement=Duration(1, 6)))
    Chord("<d' a'>8 * 1/3")        Timespan(Offset((1, 4)), Offset((1, 4), displacement=Duration(1, 24)))
    Note("b'8 * 1/3")              Timespan(Offset((1, 4), displacement=Duration(1, 24)), Offset((1, 4), displacement=Duration(1, 12)))
    Note("c''8 * 1/3")             Timespan(Offset((1, 4), displacement=Duration(1, 12)), Offset((1, 4), displacement=Duration(1, 8)))
    Note("b'8 * 1/3")              Timespan(Offset((1, 4), displacement=Duration(1, 8)), Offset((1, 4), displacement=Duration(1, 6)))
    Voice("d'4 e'4", name='MusicVoice') Timespan(Offset((1, 4)), Offset((3, 4)))
    Note("d'4")                    Timespan(Offset((1, 4), displacement=Duration(1, 6)), Offset((1, 2)))
    Note("e'4")                    Timespan(Offset((1, 2)), Offset((3, 4)))
    Note("f'4")                    Timespan(Offset((3, 4)), Offset((1, 1)))

"""


def grace_corner_cases():
    """
    Read these module-level examples.
    """
    pass



import dataclasses
import typing

from ply import lex

from .. import exceptions
from .base import Parser


@dataclasses.dataclass(slots=True)
class Scheme:
    """
    Abjad model of Scheme code.
    """

    value: typing.Any = None


class SchemeParser(Parser):
    """
    ``SchemeParser`` mimics how LilyPond's embedded Scheme parser behaves.

    It parses a single Scheme expression and then stops, by raising a
    ``SchemeParserFinishedError``.

    The parsed expression and its exact length in characters are cached on the
    ``SchemeParser`` instance.

    It is intended to be used only in conjunction with ``LilyPondParser``.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "cursor",
        "cursor_end",
        "expression_depth",
        "result",
        "string_accumulator",
    )

    ### INITIALIZER ###

    def __init__(self, debug=False):
        self.cursor = None
        self.cursor_end = None
        self.expression_depth = None
        self.result = None
        self.string_accumulator = None
        Parser.__init__(self, debug=debug)

    ### PRIVATE METHODS ###

    def _setup(self):
        self.expression_depth = 0
        self.cursor = 0

    ### LEX SETUP ###

    A = r"[A-Za-z]"
    N = r"[0-9]"
    DIGIT = rf"{N}"
    UNSIGNED = rf"{N}+"
    HEX = r"(X|x)[A-Fa-f0-9]+"
    INT = rf"(-?{UNSIGNED})"
    REAL = r"(({}\.{}*)|(-?\.{}+))".format(INT, N, N)
    INITIAL = r"({}|!|\$|%|&|\*|/|<|>|\?|~|_|\^|:|=)".format(A)
    SUBSEQUENT = r"({}|{}|\.|\+|-)".format(INITIAL, N)
    # IDENTIFIER = r'({}{}*|\+|-|\.\.\.)'.format(INITIAL, SUBSEQUENT)

    # has been rewritten to prevent Sphinx from complaining that it looks like bad ReST
    IDENTIFIER = (
        r"([A-Za-z!\$%&\*/<>\?~_\^:=][A-Za-z0-9!\$%&\*/<>\?~_\^:=\.\+-]*|[\+-]|\.\.\.)"
    )

    states = (("quote", "exclusive"),)

    tokens = (
        "AMPERSAND",
        "ASTERISK",
        "BOOLEAN",
        "CARAT",
        "COLON",
        "DECIMAL",
        "DOLLAR",
        "EQUALS",
        "EXCLAMATION",
        "HASH",
        "HEXADECIMAL",
        "IDENTIFIER",
        "INTEGER",
        "L_CARAT",
        "L_PAREN",
        "MINUS",
        "PERCENT",
        "PERIOD",
        "PLUS",
        "QUESTION",
        "QUOTE",
        "R_CARAT",
        "R_PAREN",
        "SLASH",
        "STRING",
        "TILDE",
        "UNDERSCORE",
    )

    ### LEX METHODS ###

    t_ignore = ""

    def t_whitespace(self, t):
        r"[ \t\r]+"
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        pass

    def t_BOOLEAN(self, t):
        r"\#(T|F|t|f)"
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        if t.value[-1].lower() == "t":
            t.value = True
        else:
            t.value = False
        return t

    @lex.TOKEN(HEX)
    def t_HEXADECIMAL(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = int(t.value[2:], 16)
        return t

    @lex.TOKEN(REAL)
    def t_DECIMAL(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = float(t.value)
        return t

    def t_HASH(self, t):
        r"\#"
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    @lex.TOKEN(INT)
    def t_INTEGER(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = int(t.value)
        return t

    def t_L_PAREN(self, t):
        r"\("
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.expression_depth += 1
        return t

    def t_R_PAREN(self, t):
        r"\)"
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    def t_quote(self, t):
        r"\" "
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.push_state("quote")
        self.string_accumulator = ""
        pass

    def t_quote_440(self, t):
        r"""\\[nt\\'"]"""
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    def t_quote_443(self, t):
        r'[^\\""]+'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    def t_quote_446(self, t):
        r"\" "
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.pop_state()
        t.type = "STRING"
        t.value = self.string_accumulator
        self.string_accumulator = ""
        return t

    def t_quote_456(self, t):
        r"."
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    @lex.TOKEN(IDENTIFIER)
    def t_IDENTIFIER(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    def t_newline(self, t):
        r"\n+"
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.lineno += t.value.count("\n")

    def t_anything(self, t):
        r"."
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        types = {
            "&": "AMPERSAND",
            "*": "ASTERISK",
            "^": "CARAT",
            ":": "COLON",
            "$": "DOLLAR",
            "=": "EQUALS",
            "!": "EXCLAMATION",
            "<": "L_CARAT",
            "-": "MINUS",
            "%": "PERCENT",
            ".": "PERIOD",
            "+": "PLUS",
            "?": "QUESTION",
            "'": "QUOTE",
            ">": "R_CARAT",
            ")": "R_PAREN",
            "/": "SLASH",
            "~": "TILDE",
            "_": "UNDERSCORE",
        }
        if t.value in types:
            t.type = types[t.value]
            return t
        else:
            # t.lexer.skip(1)
            pass

    def t_error(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        if self.debug:
            print(f"SchemeParser-{id(self)}: Illegal character {t.value[0]!r}")
        # t.lexer.skip(1)

    t_quote_error = t_error
    t_quote_ignore = t_ignore

    ### YACC SETUP ###

    start = "program"

    ### YACC METHODS ###

    ### program ###

    """<program> : <form>*
    """

    def p_program__forms(self, p):
        """
        program : forms
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_forms__EMPTY(self, p):
        """
        forms :
        """
        p[0] = []

    def p_forms__forms__form(self, p):
        """
        forms : forms form
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1] + [p[2]]

    ### form ###

    """<form> : <definition> | <expression>
    """

    def p_form__expression(self, p):
        """
        form : expression
        """
        # print 'form : expression'
        # print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        self.result = p[0]
        self.cursor_end = p.slice[0].cursor_end
        if self.debug:
            print("PARSED {!r}".format(self.lexer.lexdata[: self.cursor_end]))
        raise exceptions.SchemeParserFinishedError

    ### definition ###

    """<definition> : <variable definition>
    |   <syntax definition>
    |   (begin <definition>*)
    |   (let-syntax (<syntax binding>*) <definition>*)
    |   (letrec-syntax (<syntax binding>*) <definition>*)
    |   <derived definition>
    """

    ### variable definition ###

    """<variable definition> : (define <variable> <expression>)
    |   (define (<variable> <variable>*) <body>)
    |   (define (<variable> <variable>* . <variable>) <body>)
    """

    ### variable ###

    """<variable> : <identifier>
    """

    def p_variable__IDENTIFIER(self, p):
        """
        variable : IDENTIFIER
        """
        # print 'variable : IDENTIFIER'
        # print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        raise Exception(p, "AAA")
        p[0] = Scheme(p[1])

    ### body ###

    """<body> : <definition>* <expression>+
    """

    ### syntax definition ###

    """<syntax definition> : (define-syntax <keyword> <transformer expression>)
    """

    ### keyword ###

    """<keyword> : <identifier>
    """

    ### syntax binding ###

    """<syntax binding> : (<keyword> <transformer expression>)
    """

    ### expression ###

    """<expression> : <constant>
    |   <variable>
    |   (quote <datum>) | ' <datum>
    |   (lambda <formals> <body>)
    |   (if <expression> <expression> <expression>) | (if <expression> <expression>)
    |   (set! <variable> <expression>)
    |   <application>
    |   (let-syntax (<syntax binding>*) <expression>+)
    |   (letrec-syntax (<syntax binding>*) <expression>+)
    |   <derived expression>
    """

    def p_expression__variable(self, p):
        """
        expression : variable
        """
        # print 'expression : variable'
        # print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_expression__QUOTE__datum(self, p):
        """
        expression : QUOTE datum
        """
        # print 'expression : QUOTE datum'
        # print p[2]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        datum = p[2]
        p[0] = Scheme(datum)

    def p_expression__constant(self, p):
        """
        expression : constant
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### constant ###

    """<constant> : <boolean> | <number> | <character> | <string>
    """

    def p_constant__boolean(self, p):
        """
        constant : boolean
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            # print 'constant : boolean'
            # print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    def p_constant__number(self, p):
        """
        constant : number
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            # print 'constant : number'
            # print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    def p_constant__string(self, p):
        """
        constant : string
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            # print 'constant : string'
            # print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    ### formals ###

    """<formals> : <variable> | (<variable>*) | (<variable>+ . <variable>)
    """

    ### application ###

    """<application> : (<expression> <expression>*)
    """

    ### identifier ###

    """<identifier> : <initial> <subsequent>* | + | - | ...
    """

    ### initial ###

    """<initial> : <letter> | ! | $ | % | & | * | / | : | < | = | > | ? | ~ | _ | ^
    """

    ### subsequent ###

    """<subsequent> : <initial> | <digit> | . | + | -
    """

    ### letter ###

    """<letter> : a | b | ... | z
    """

    ### digit ###

    """<digit> : 0 | 1 | ... | 9
    """

    ### datum ###

    """<datum> : <boolean> | <number> | <character> | <string> | <symbol> | <list> | <vector>
    """

    def p_datum__constant(self, p):
        """
        datum : constant
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_datum__symbol(self, p):
        """
        datum : symbol
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    # def p_datum__boolean(self, p):
    #    """datum : boolean"""
    #    p[0] = p[1]

    # def p_datum__number(self, p):
    #    """datum : number"""
    #    p[0] = p[1]

    # def p_datum__string(self, p):
    #    """datum : string"""
    #    p[0] = p[1]

    def p_datum__list(self, p):
        """
        datum : list
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_datum__vector(self, p):
        """
        datum : vector
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_data__EMPTY(self, p):
        """
        data :
        """
        p[0] = []

    def p_data__data__datum(self, p):
        """
        data : data datum
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1] + [p[2]]

    ### boolean ###

    """<boolean> : #t | #f
    """

    def p_boolean__BOOLEAN(self, p):
        """
        boolean : BOOLEAN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### number ###

    """<number> : <num 2> | <num 8> | <num 10> | <num 16>
    """

    def p_number__DECIMAL(self, p):
        """
        number : DECIMAL
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_number__HEXADECIMAL(self, p):
        """
        number : HEXADECIMAL
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_number__INTEGER(self, p):
        """
        number : INTEGER
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### character ###

    r"""<character> : #\ <any character> | #\newline | #\space
    """

    ### string ###

    """<string> : " <string character>* "
    """

    def p_string__STRING(self, p):
        """
        string : STRING
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### string character ###

    r"""<string character> : \" | \\ | <any character other than " or \>
    """

    ### symbol ###

    """<symbol> : <identifier>
    """

    def p_symbol__IDENTIFIER(self, p):
        """
        symbol : IDENTIFIER
        """
        # print 'symbol : IDENTIFIER'
        # print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### vector ###

    """<vector> : #(<datum>*)
    """

    def p_vector__HASH__L_PAREN__data__R_PAREN(self, p):
        """
        vector : HASH L_PAREN data R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[3]
        self.expression_depth -= 1

    ### list ###

    """<list> : (<datum>*) | (<datum>+ . <datum>) | <abbreviation>
    """

    def p_list__L_PAREN__data__R_PAREN(self, p):
        """
        list : L_PAREN data R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[2]
        self.expression_depth -= 1

    def p_list__L_PAREN__data__datum__PERIOD__datum__R_PAREN(self, p):
        """
        list : L_PAREN data datum PERIOD datum R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        self.expression_depth -= 1

    ### abbreviation ###

    """<abbreviation> : ' <datum> | ` <datum> | , <datum> | ,@ <datum>
    """

    ### error ###

    def p_error(self, p):
        if p:
            if self.debug:
                print("SchemeParser-{}: Syntax error at {!r}".format(id(self), p.value))
            self._parser.errok()
        else:
            if self.debug:
                print("SchemeParser-{}: Syntax error at EOF".format(id(self)))

    ### PUBLIC PROPERTIES ###

    @property
    def lexer_rules_object(self):
        """
        Lexer rules object of Scheme parser.
        """
        return self

    @property
    def parser_rules_object(self):
        """
        Parser rules object of Scheme parser.
        """
        return self



import pytest

import abjad


def test_get_bar_line_crossing_01():
    """
    Works with partial.
    """

    staff = abjad.Staff("c'8 d'8 e'4 f'8")
    abjad.Score([staff], name="Score")
    time_signature = abjad.TimeSignature((2, 8), partial=abjad.Duration(1, 8))
    abjad.attach(time_signature, staff[0])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \partial 8
            \time 2/8
            c'8
            d'8
            e'4
            f'8
        }
        """
    )

    assert not abjad.get.bar_line_crossing(staff[0])
    assert not abjad.get.bar_line_crossing(staff[1])
    assert abjad.get.bar_line_crossing(staff[2])
    assert not abjad.get.bar_line_crossing(staff[3])


def test_get_bar_line_crossing_02():
    """
    Works when no explicit time signature is abjad.attached.
    """

    staff = abjad.Staff("c'2 d'1 e'2")

    assert not abjad.get.bar_line_crossing(staff[0])
    assert abjad.get.bar_line_crossing(staff[1])
    assert not abjad.get.bar_line_crossing(staff[2])


def test_get_duration_01():
    """
    Container duration in seconds equals sum of leaf durations in seconds.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    score = abjad.Score([staff])
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 38)
    abjad.attach(mark, staff[0])
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 42)
    abjad.attach(mark, staff[2])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \tempo 4=38
                c'8
                d'8
                \tempo 4=42
                e'8
                f'8
            }
        >>
        """
    )

    assert abjad.get.duration(score, in_seconds=True) == abjad.Duration(400, 133)


def test_get_duration_02():
    """
    Container can not calculate duration in seconds without metronome mark.
    """

    container = abjad.Container("c'8 d'8 e'8 f'8")
    with pytest.raises(Exception):
        abjad.get.duration(container, in_seconds=True)


def test_get_duration_03():
    """
    Clock duration equals duration divide by effective tempo.
    """

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    abjad.Score([staff])
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 38)
    abjad.attach(mark, staff[0])
    mark = abjad.MetronomeMark(abjad.Duration(1, 4), 42)
    abjad.attach(mark, staff[2])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \tempo 4=38
            c'8
            d'8
            \tempo 4=42
            e'8
            f'8
        }
        """
    )

    assert abjad.get.duration(staff[0], in_seconds=True) == abjad.Duration(15, 19)
    assert abjad.get.duration(staff[1], in_seconds=True) == abjad.Duration(15, 19)
    assert abjad.get.duration(staff[2], in_seconds=True) == abjad.Duration(5, 7)
    assert abjad.get.duration(staff[3], in_seconds=True) == abjad.Duration(5, 7)


def test_get_duration_04():
    """
    Clock duration can not calculate without metronome mark.
    """

    note = abjad.Note("c'4")
    with pytest.raises(Exception):
        abjad.get.duration(note, in_seconds=True)


def test_get_has_effective_indicator_01():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    abjad.attach("foo", staff[2], context="Staff")

    assert not abjad.get.has_effective_indicator(staff, str)
    assert not abjad.get.has_effective_indicator(staff[0], str)
    assert not abjad.get.has_effective_indicator(staff[1], str)
    assert abjad.get.has_effective_indicator(staff[2], str)
    assert abjad.get.has_effective_indicator(staff[3], str)


def test_get_has_indicator_01():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    abjad.attach("foo", staff[0])

    assert not abjad.get.has_indicator(staff, str)
    assert abjad.get.has_indicator(staff[0], str)
    assert not abjad.get.has_indicator(staff[1], str)
    assert not abjad.get.has_indicator(staff[2], str)
    assert not abjad.get.has_indicator(staff[3], str)


def test_get_has_indicator_02():
    staff = abjad.Staff("c'2 d'2")
    articulation = abjad.Articulation("staccato")
    abjad.attach(articulation, staff[0])

    assert abjad.get.has_indicator(staff[0], abjad.Articulation)
    assert not abjad.get.has_indicator(staff[1], abjad.Duration)


def test_get_has_indicator_03():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    command = abjad.LilyPondLiteral(r"\break", site="after")
    abjad.attach(command, staff[-1])

    assert not abjad.get.has_indicator(staff[0], abjad.LilyPondLiteral)
    assert not abjad.get.has_indicator(staff[1], abjad.LilyPondLiteral)
    assert not abjad.get.has_indicator(staff[2], abjad.LilyPondLiteral)
    assert abjad.get.has_indicator(staff[3], abjad.LilyPondLiteral)


def test_get_has_indicator_04():
    staff = abjad.Staff("c'2 d'2")
    stem_tremolo = abjad.StemTremolo(16)
    abjad.attach(stem_tremolo, staff[0])

    assert abjad.get.has_indicator(staff[0], abjad.StemTremolo)
    assert not abjad.get.has_indicator(staff[1], abjad.StemTremolo)


def test_get_has_indicator_05():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    abjad.Score([staff], name="Score")
    time_signature = abjad.TimeSignature((4, 8))
    abjad.attach(time_signature, staff[0])

    assert abjad.get.has_indicator(staff[0], abjad.TimeSignature)
    assert not abjad.get.has_indicator(staff[1])
    assert not abjad.get.has_indicator(staff[2])
    assert not abjad.get.has_indicator(staff[3])
    assert not abjad.get.has_indicator(staff)


def test_get_indicators_01():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.slur(voice[:])
    command_1 = abjad.LilyPondLiteral(r"\slurDotted", site="before")
    abjad.attach(command_1, voice[0])
    command_2 = abjad.LilyPondLiteral(r"\slurUp", site="before")
    abjad.attach(command_2, voice[0])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            \slurDotted
            \slurUp
            c'8
            (
            d'8
            e'8
            f'8
            )
        }
        """
    ), abjad.lilypond(voice)

    indicators = abjad.get.indicators(voice[0], abjad.LilyPondLiteral)
    assert command_1 in indicators
    assert command_2 in indicators
    assert len(indicators) == 2


def test_get_indicators_02():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff([voice])
    clef = abjad.Clef("treble")
    abjad.attach(clef, voice[0])
    dynamic = abjad.Dynamic("p")
    abjad.attach(dynamic, voice[0])

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                \clef "treble"
                c'8
                \p
                d'8
                e'8
                f'8
            }
        }
        """
    ), abjad.lilypond(staff)

    indicators = abjad.get.indicators(voice[0])
    assert len(indicators) == 2


def test_get_indicators_03():
    note = abjad.Note("c'4")
    stem_tremolo = abjad.StemTremolo(16)
    abjad.attach(stem_tremolo, note)
    stem_tremolos = abjad.get.indicators(note, abjad.StemTremolo)

    assert stem_tremolos[0] is stem_tremolo


def test_get_markup_01():
    chord = abjad.Chord([-11, 2, 5], (1, 4))
    up_markup = abjad.Markup(r"\markup UP")
    abjad.attach(up_markup, chord, direction=abjad.UP)
    down_markup = abjad.Markup(r"\markup DOWN")
    abjad.attach(down_markup, chord, direction=abjad.DOWN)
    found_markup = abjad.get.markup(chord, direction=abjad.DOWN)
    assert found_markup == [down_markup]


def test_get_markup_02():
    chord = abjad.Chord([-11, 2, 5], (1, 4))
    up_markup = abjad.Markup(r"\markup UP")
    abjad.attach(up_markup, chord, direction=abjad.UP)
    down_markup = abjad.Markup(r"\markup DOWN")
    abjad.attach(down_markup, chord, direction=abjad.DOWN)
    found_markup = abjad.get.markup(chord, direction=abjad.UP)
    assert found_markup == [up_markup]


def test_get_staff_01():
    """
    Staff changes work on the first note of a staff.
    """

    staff_group = abjad.StaffGroup(
        [abjad.Staff("c'8 d'8 e'8 f'8"), abjad.Staff("c'8 d'8 e'8 f'8")]
    )
    staff_group.lilypond_type = "PianoStaff"
    staff_group.simultaneous = True
    staff_group[0].name = "RH"
    staff_group[1].name = "LH"
    staff_change = abjad.StaffChange("LH")
    abjad.attach(staff_change, staff_group[0][0])

    assert abjad.lilypond(staff_group) == abjad.string.normalize(
        r"""
        \new PianoStaff
        <<
            \context Staff = "RH"
            {
                \change Staff = LH
                c'8
                d'8
                e'8
                f'8
            }
            \context Staff = "LH"
            {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        """
    )

    assert abjad.wf.wellformed(staff_group)
    assert abjad.get.effective_staff(staff_group[0][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][2]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][3]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][2]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][3]) is staff_group[1]


def test_get_staff_02():
    """
    Staff changes work on middle notes of a staff.
    """

    staff_group = abjad.StaffGroup(
        [abjad.Staff("c'8 d'8 e'8 f'8"), abjad.Staff("c'8 d'8 e'8 f'8")]
    )
    staff_group.lilypond_type = "PianoStaff"
    staff_group.simultaneous = True
    staff_group[0].name = "RH"
    staff_group[1].name = "LH"
    staff_change = abjad.StaffChange("LH")
    abjad.attach(staff_change, staff_group[0][0])
    staff_change = abjad.StaffChange("RH")
    abjad.attach(staff_change, staff_group[0][2])

    assert abjad.lilypond(staff_group) == abjad.string.normalize(
        r"""
        \new PianoStaff
        <<
            \context Staff = "RH"
            {
                \change Staff = LH
                c'8
                d'8
                \change Staff = RH
                e'8
                f'8
            }
            \context Staff = "LH"
            {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        """
    )

    assert abjad.wf.wellformed(staff_group)
    assert abjad.get.effective_staff(staff_group[0][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][2]) is staff_group[0]
    assert abjad.get.effective_staff(staff_group[0][3]) is staff_group[0]
    assert abjad.get.effective_staff(staff_group[1][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][2]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][3]) is staff_group[1]


def test_get_staff_03():
    """
    Staff changes work on the last note of a staff.
    """

    staff_group = abjad.StaffGroup(
        [abjad.Staff("c'8 d'8 e'8 f'8"), abjad.Staff("c'8 d'8 e'8 f'8")]
    )
    staff_group.lilypond_type = "PianoStaff"
    staff_group.simultaneous = True
    staff_group[0].name = "RH"
    staff_group[1].name = "LH"
    staff_change = abjad.StaffChange("LH")
    abjad.attach(staff_change, staff_group[0][-1])

    assert abjad.lilypond(staff_group) == abjad.string.normalize(
        r"""
        \new PianoStaff
        <<
            \context Staff = "RH"
            {
                c'8
                d'8
                e'8
                \change Staff = LH
                f'8
            }
            \context Staff = "LH"
            {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        """
    )

    assert abjad.wf.wellformed(staff_group)


def test_get_staff_04():
    """
    Redudant staff changes are allowed.
    """

    staff_group = abjad.StaffGroup(
        [abjad.Staff("c'8 d'8 e'8 f'8"), abjad.Staff("c'8 d'8 e'8 f'8")]
    )
    staff_group.lilypond_type = "PianoStaff"
    staff_group.simultaneous = True
    staff_group[0].name = "RH"
    staff_group[1].name = "LH"
    staff_change = abjad.StaffChange("LH")
    abjad.attach(staff_change, staff_group[0][0])
    staff_change = abjad.StaffChange("LH")
    abjad.attach(staff_change, staff_group[0][1])

    assert abjad.lilypond(staff_group) == abjad.string.normalize(
        r"""
        \new PianoStaff
        <<
            \context Staff = "RH"
            {
                \change Staff = LH
                c'8
                \change Staff = LH
                d'8
                e'8
                f'8
            }
            \context Staff = "LH"
            {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        """
    )

    assert abjad.wf.wellformed(staff_group)
    assert abjad.get.effective_staff(staff_group[0][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][2]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[0][3]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][0]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][1]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][2]) is staff_group[1]
    assert abjad.get.effective_staff(staff_group[1][3]) is staff_group[1]
